from typing import List, Set, Dict, Tuple
from dotenv import dotenv_values, set_key, unset_key, parser
from ..contracts_generated import AbstraLibApiEditorEnvVarsModel
from ..settings import Settings
from .project.project import ProjectRepository
from ..utils.code import function_called_args, subscript_called_args
import ast
from pathlib import Path


class EnvVarsRepository:
    @staticmethod
    def get_env_var_path() -> Path:
        return Settings.root_path / ".env"

    @staticmethod
    def list() -> List[AbstraLibApiEditorEnvVarsModel]:
        values = dotenv_values(EnvVarsRepository.get_env_var_path())
        return [
            AbstraLibApiEditorEnvVarsModel.from_dict(
                {
                    "key": k,
                    "value": v,
                }
            )
            for k, v in values.items()
        ]

    @staticmethod
    def set(env_var: AbstraLibApiEditorEnvVarsModel) -> AbstraLibApiEditorEnvVarsModel:
        set_key(EnvVarsRepository.get_env_var_path(), env_var.key, env_var.value)
        return env_var

    @staticmethod
    def unset(
        env_var: AbstraLibApiEditorEnvVarsModel,
    ) -> AbstraLibApiEditorEnvVarsModel:
        unset_key(EnvVarsRepository.get_env_var_path(), env_var.key)
        return env_var

    @staticmethod
    def check() -> bool:
        env_file_path = EnvVarsRepository.get_env_var_path()
        if not env_file_path.exists():
            return True
        with open(env_file_path) as f:
            try:
                for it in parser.parse_stream(f):
                    if it.error:
                        return False
                return True
            except parser.Error as e:
                print(e)
                return False

    @classmethod
    def get_env_vars_in_code(cls) -> Dict[str, Set[Tuple[Path, ast.expr]]]:
        used_env_vars: Dict[str, Set[Tuple[Path, ast.expr]]] = {}
        project = ProjectRepository.load()

        for python_file in project.project_files:
            try:
                function_calls = function_called_args(
                    python_file.read_text(encoding="utf-8"), ["os"], "getenv"
                )
                subscript_calls = subscript_called_args(
                    python_file.read_text(encoding="utf-8"), ["os"], "environ"
                )

                if function_calls is not None:
                    for function_call in function_calls:
                        for arg in function_call:
                            if isinstance(arg, ast.Constant) and isinstance(
                                arg.value, str
                            ):
                                used_env_vars.setdefault(arg.value, set()).add(
                                    (python_file, arg)
                                )
                if subscript_calls is not None:
                    for subscript_call in subscript_calls:
                        for arg in subscript_call:
                            if isinstance(arg, ast.Constant) and isinstance(
                                arg.value, str
                            ):
                                used_env_vars.setdefault(arg.value, set()).add(
                                    (python_file, arg)
                                )
            except SyntaxError:
                continue
            except FileNotFoundError:
                continue

        return used_env_vars


def env_vars_repository_factory() -> EnvVarsRepository:
    return EnvVarsRepository()



import webbrowser
from typing import List
from pathlib import Path

from ...utils.file import traverse_code
from ..linter import LinterRule, LinterIssue, LinterFix
from ...repositories.project.project import ProjectRepository


class OpenBrokenFile(LinterFix):
    file: Path

    def __init__(self, file: Path) -> None:
        self.label = f"Open {file}"
        self.file = file

    def fix(self):
        webbrowser.open(self.file.absolute().as_uri())


class SyntaxErrorFound(LinterIssue):
    def __init__(self, error: SyntaxError) -> None:
        self.label = str(error)
        if error.filename is not None:
            self.fixes = [OpenBrokenFile(Path(error.filename))]


class SyntaxErrors(LinterRule):
    label = "Syntax errors"
    type = "bug"
    error = None

    def find_issues(self) -> List[LinterIssue]:
        project = ProjectRepository.load()
        for entrypoint in project.iter_entrypoints():
            try:
                for _ in traverse_code(entrypoint, raise_on_syntax_errors=True):
                    pass
            except SyntaxError as e:
                self.error = e
                self.error.filename = str(entrypoint)
                return [SyntaxErrorFound(e)]
        return []



import json, requests
from pathlib import Path
from datetime import datetime
from packaging.version import Version

from .dot_abstra import CACHED_VERSIONS_DIR


EXPIRE_PERIOD = 60 * 60 * 4  # 4 hours
TIMEOUT = 5


def get_local_package_version(package_name="abstra"):
    import pkg_resources

    return parse_version(pkg_resources.get_distribution(package_name).version)


def parse_version(version: str):
    return Version(version)


def get_cached_latest_version(root_path: Path, package_name="abstra"):
    cached_file = root_path / CACHED_VERSIONS_DIR / f"{package_name}.json"

    if cached_file.exists():
        try:
            with open(cached_file, "r") as f:
                cached_version = json.loads(f.readline())

            created_at = cached_version["created_at"]
            version = cached_version["version"]
        except:
            created_at = None
            version = None

        if (
            created_at is not None
            and isinstance(created_at, float)
            and version is not None
            and datetime.utcnow().timestamp() - created_at < EXPIRE_PERIOD
        ):
            return Version(version)

    response = requests.get(
        f"https://pypi.org/pypi/{package_name}/json", timeout=TIMEOUT
    )
    latest_version = response.json()["info"]["version"]

    update_cached_latest_version(root_path, latest_version, package_name)

    return Version(latest_version)


def update_cached_latest_version(
    root_path: Path, version: Version, package_name="abstra"
):
    cached_file = root_path / CACHED_VERSIONS_DIR / f"{package_name}.json"

    if not cached_file.parent.exists():
        cached_file.parent.mkdir(parents=True)

    with open(cached_file, "w") as f:
        json.dump(
            {
                "version": str(version),
                "created_at": datetime.utcnow().timestamp(),
            },
            f,
        )
